/* XCC (mips): Experimental C-subset Compiler.
  Copyright (c) 2002-2008, gondow@cs.titech.ac.jp, All rights reserved.
  $Id: codegen-mips.c,v 1.16 2009/03/27 10:17:08 gondow Exp gondow $ */ 
/* ---------------------------------------------------------------------- */
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "xcc.h"
#include "misc.h"
#include "AST.h"
#include "type.h"
#include "symbol.h"
#include "codegen.h"

static char *func_name;
static int   frame_size, total_arg_size, total_local_size;

static void emit_code (struct AST *ast, char *fmt, ...);
static int push_reg (void);
static int pop_reg  (void);
static int top_reg  (void);
static void codegen_exp_unary (struct AST *ast);
static void codegen_pointer_offset (struct AST *ast, int size);
static void codegen_exp_add (struct AST *ast);
static void codegen_exp_sub (struct AST *ast);
static void codegen_exp_binop1 (struct AST *ast);
static void codegen_exp_binop2 (struct AST *ast);
static void codegen_lvalue (struct AST *ast);
static void codegen_exp_funcall (struct AST *ast);
static void codegen_exp_id (struct AST *ast);
static void codegen_exp (struct AST *ast);
static char *create_ctrl_label (void);
static void codegen_stmt (struct AST *ast_stmt);
static void codegen_begin_block (struct AST *ast);
static void codegen_end_block (void);
static void codegen_block (struct AST *ast_block);
static void codegen_begin_function (struct AST *ast);
static void codegen_end_function (void);
static void codegen_func (struct AST *ast);
static void codegen_dec (struct AST *ast);

enum { REG_NUM = 9 };
struct REG {
    char *name [REG_NUM];
    int   index; /* name[0] through name[index] are used. */
};
static struct REG reg = {
    {"$s0", "$s1", "$s2", "$s3", "$s4", "$s5", "$s6", "$s7"},
    -1
};

/* ---------------------------------------------------------------------- */

static void
emit_code (struct AST *ast, char *fmt, ...)
{
    va_list  argp;
    va_start (argp, fmt);
    vfprintf (xcc_out, fmt, argp);
    va_end   (argp);

    /* the argument 'ast' can be used for debug purpose */
}

#ifdef XCC_VIS
#include "vis/emit_code.h"
#endif

static int
push_reg (void)
{
    reg.index++;
    if (reg.index >= REG_NUM) {
	fprintf (stderr, "reg stack overflow\n");
    }
    return reg.index;
}

static int
pop_reg (void)
{
    int old_index = reg.index;
    reg.index--;
    if (reg.index < -1) {
	fprintf (stderr, "reg stack underflow\n");
    }
    return old_index;
}

static int
top_reg (void)
{
    return reg.index;
}

static void
codegen_exp_unary (struct AST *ast)
{
    int reg1, reg2;

    if (!strcmp (ast->child [0]->ast_type, "AST_unary_operator_address")) {
	codegen_lvalue (ast->child [1]);
    } else if (!strcmp (ast->child [0]->ast_type, "AST_unary_operator_deref")) {
	codegen_exp (ast->child [1]);
	reg1 = pop_reg ();
	reg2 = push_reg ();
	emit_code (ast->child [0], "\tlw      %s, 0(%s)\n",
                   reg.name [reg2], reg.name [reg2]);
    } else if (!strcmp (ast->child [0]->ast_type, "AST_unary_operator_plus")) {
	codegen_exp (ast->child [1]);
    } else if (!strcmp (ast->child [0]->ast_type, "AST_unary_operator_minus")) {
	codegen_exp (ast->child [1]);
	reg1 = top_reg ();
	emit_code (ast->child [0], "\tsubu    %s, $0, %s\n",
                   reg.name [reg1], reg.name [reg1]);
    } else if (!strcmp (ast->child [0]->ast_type, "AST_unary_operator_negative")) {
	codegen_exp (ast->child [1]);
	reg1 = top_reg ();
	emit_code (ast->child [0], "\tsltu %s, %s, 1\n",
                   reg.name [reg1], reg.name [reg1]);
    } else {
        assert (0);
    }
}

static void
codegen_pointer_offset (struct AST *ast, int size)
{
    int reg1, reg2, reg3;

    reg1 = push_reg ();
    emit_code (ast, "\tli      %s, %d\n", reg.name [reg1], size);
    reg2 = pop_reg ();
    reg3 = top_reg ();
    emit_code (ast, "\tmult    %s, %s\n", reg.name [reg2], reg.name [reg3]);
    emit_code (ast, "\tmflo    %s\n",     reg.name [reg3]);
}

static void
codegen_exp_add (struct AST *ast)
{
    struct AST *ast0, *ast1;
    int reg1, reg2, reg3;

    ast0 = ast->child [0];
    ast1 = ast->child [1];
    assert (   (ast0->type->kind != TYPE_KIND_POINTER)
	    || (ast1->type->kind != TYPE_KIND_POINTER));

    codegen_exp (ast0);
    if (ast1->type->kind == TYPE_KIND_POINTER)
        codegen_pointer_offset (ast, ast1->type->u.t_pointer.type->size);

    codegen_exp (ast1);
    if (ast0->type->kind == TYPE_KIND_POINTER)
        codegen_pointer_offset (ast, ast0->type->u.t_pointer.type->size);

    reg1 = pop_reg ();
    reg2 = pop_reg ();
    reg3 = push_reg ();
    emit_code (ast, "\tadd     %s, %s, %s\n",
	       reg.name [reg2], reg.name [reg1], reg.name [reg3]);
}

static void
codegen_exp_sub (struct AST *ast)
{
    struct AST *ast0, *ast1;
    struct Type *type0, *type1;
    int size;
    int reg1, reg2, reg3;

    ast0 = ast->child [0];
    ast1 = ast->child [1];
    type0 = ast0->type;
    type1 = ast1->type;

    assert ((type0->kind == TYPE_KIND_POINTER)
            || (type1->kind != TYPE_KIND_POINTER));

    codegen_exp (ast0);
    codegen_exp (ast1);
    if ((type0->kind == TYPE_KIND_POINTER)
	&& (type1->kind != TYPE_KIND_POINTER)) {
	size = type0->u.t_pointer.type->size;
        reg1 = push_reg ();
        emit_code (ast, "\tli      %s, %d\n", reg.name [reg1], size);
        reg2 = pop_reg ();
        reg3 = top_reg ();
        emit_code (ast, "\tmult    %s, %s\n", reg.name [reg2], reg.name [reg3]);
        emit_code (ast, "\tmflo    %s\n", reg.name [reg3]);

    }
    reg1 = pop_reg ();
    reg2 = pop_reg ();
    reg3 = push_reg ();
    emit_code (ast, "\tsubu    %s, %s, %s\n",
               reg.name [reg3], reg.name [reg2], reg.name [reg1]);

    if ((type0->kind == TYPE_KIND_POINTER)
	&& (type1->kind == TYPE_KIND_POINTER)) {
	assert (type0->size == type1->size);
	reg1 = push_reg ();
        emit_code (ast, "\tli      %s, %d\n", reg.name [reg1], type0->size);
	reg2 = pop_reg ();
        reg3 = top_reg ();
	emit_code (ast, "\tdiv     %s, %s\n", reg.name [reg3], reg.name [reg2]);
        emit_code (ast, "\tmflo    %s\n", reg.name [reg3]);
    }
}

static void
codegen_exp_binop1 (struct AST *ast)
{
    int reg1, reg2;
    char *label1, *label2;

    if (!strcmp (ast->ast_type, "AST_expression_lor")) {
	label1 = create_ctrl_label ();
	label2 = create_ctrl_label ();

        codegen_exp (ast->child [0]);
        reg1 = pop_reg ();
        emit_code (ast, "\tbne     %s, $0, %s\n", reg.name [reg1], label1);
	emit_code (ast, "\tnop\n");

        codegen_exp (ast->child [1]);
        reg1 = pop_reg ();
        emit_code (ast, "\tbne     %s, $0, %s\n", reg.name [reg1], label1);
	emit_code (ast, "\tnop\n");

        reg1 = push_reg ();
        emit_code (ast, "\tj       %s\n", label2);
	emit_code (ast, "\tnop\n");
        emit_code (ast, "%s:\n", label1);
        emit_code (ast, "\tli      %s, 1\n", reg.name [reg1]);
        emit_code (ast, "%s:\n", label2);
    } else if (!strcmp (ast->ast_type, "AST_expression_land")) {
	label1 = create_ctrl_label ();

        codegen_exp (ast->child [0]);
        reg1 = pop_reg ();
        emit_code (ast, "\tbeq     %s, $0, %s\n", reg.name [reg1], label1);
	emit_code (ast, "\tnop\n");

        codegen_exp (ast->child [1]);
        reg1 = pop_reg ();
        emit_code (ast, "\tbeq     %s, $0, %s\n", reg.name [reg1], label1);
	emit_code (ast, "\tnop\n");
        reg1 = push_reg ();
        emit_code (ast, "\tli      %s, 1\n", reg.name [reg1]);
        emit_code (ast, "%s:\n", label1);
    } else if (!strcmp (ast->ast_type, "AST_expression_mul")) {
        codegen_exp (ast->child [0]);
        codegen_exp (ast->child [1]);
        reg1 = pop_reg ();
        reg2 = top_reg ();
        emit_code (ast, "\tmult    %s, %s\n",
                   reg.name [reg2], reg.name [reg1]);
        emit_code (ast, "\tmflo    %s\n", reg.name [reg2]);
    } else if (!strcmp (ast->ast_type, "AST_expression_div")) {
        codegen_exp (ast->child [0]);
        codegen_exp (ast->child [1]);
        reg1 = pop_reg ();
        reg2 = top_reg ();
	emit_code (ast, "\tdiv     %s, %s\n",
                   reg.name [reg2], reg.name [reg1]);
        emit_code (ast, "\tmflo    %s\n", reg.name [reg2]);
    } else {
        assert (0);
    }
}

static void
codegen_exp_binop2 (struct AST *ast)
{
    int reg1, reg2,reg3;

    codegen_exp (ast->child [0]);
    codegen_exp (ast->child [1]);
    reg1 = pop_reg ();
    reg2 = pop_reg ();
    reg3 = push_reg ();
    if (!strcmp (ast->ast_type, "AST_expression_eq")) {
        emit_code (ast, "\txor     %s, %s, %s\n",
                   reg.name [reg3], reg.name [reg2], reg.name [reg1]);
        emit_code (ast, "\tsltu    %s, %s, 1\n",
                   reg.name [reg3], reg.name [reg3]);
    } else if (!strcmp (ast->ast_type, "AST_expression_less")) {
        emit_code (ast, "\tslt     %s, %s, %s\n",
                   reg.name [reg3], reg.name [reg2], reg.name [reg1]);
    } else {
        assert (0);
    }
}

static void
codegen_lvalue (struct AST *ast)
{
    int reg1;
    int offset;
    struct Symbol *sym;

    if (!strcmp (ast->ast_type, "AST_expression_id")) {
	sym = sym_lookup (ast->child [0]->u.id);
	assert (sym != NULL);
	reg1 = push_reg ();
	switch (sym->name_space) {
	case NS_LOCAL:
            offset = total_arg_size + sym->offset + 8;
            /* 8 bytes for %gp + 4byte alignment */
            emit_code (ast, "\taddiu   %s, $fp, %d\t# %s, %d\n",
                       reg.name [reg1], offset, sym->name, sym->offset);
	    break;
	case NS_ARG:
            offset = frame_size + sym->offset;
            emit_code (ast, "\taddiu   %s, $fp, %d\t# %s, %d\n",
                       reg.name [reg1], offset, sym->name, sym->offset);
	    break;
	case NS_GLOBAL:
            emit_code (ast, "\tlw      %s, %%got(%s)($gp)\n",
                       reg.name [reg1], sym->name);
	    break;
	case NS_LABEL: /* falling through */
	default: assert (0); break;
	}
    } else if (!strcmp (ast->ast_type, "AST_expression_unary")) {
        if (!strcmp (ast->child [0]->ast_type, "AST_unary_operator_deref")) {
	    codegen_exp (ast->child [1]);
        } else {
            assert (0);
	}
    } else if (!strcmp (ast->ast_type, "AST_expression_paren")) {
	codegen_lvalue (ast->child [0]);
    } else {
        assert (0);
    }
}

static void
codegen_exp_funcall (struct AST *ast_func)
{
    int i, arg_num;
    int reg1, reg2;
    struct AST *ast, *ast_exp;

    assert (!strcmp (ast_func->ast_type, "AST_expression_funcall1")
	    || !strcmp (ast_func->ast_type, "AST_expression_funcall2"));

    /* processing arguments (funcall2 has no arguments) */
    if (!strcmp (ast_func->ast_type, "AST_expression_funcall1")) {
	ast = search_AST_bottom (ast_func->child [1],
				 "AST_argument_expression_list_single",
				 &arg_num);
	for (i = 0; i < arg_num; i++) {
	    if (!strcmp (ast->ast_type,
                         "AST_argument_expression_list_single")) {
		ast_exp = ast->child [0];
            } else if (!strcmp (ast->ast_type,
                                "AST_argument_expression_list_pair")) {
		ast_exp = ast->child [1];
            } else {
                assert (0);
	    }
	    codegen_exp (ast_exp);
	    ast = ast->parent;
	}

	for (i = arg_num - 1; i >= 0; i--) {
	    reg1 = pop_reg ();
	    if (i < 4)
		emit_code (ast_func, "\tmove    $a%d, %s \t# arg #%d\n",
                           i, reg.name [reg1], i);
            emit_code (ast_func, "\tsw      %s, %d($fp) \t# arg #%d\n",
                       reg.name [reg1], i * 4, i);
	}
    }

    codegen_exp (ast_func->child [0]);
    reg1 = pop_reg ();
    emit_code (ast_func, "\tsw      $v0, %d($fp)\t# save return value\n",
               frame_size - 48);
    /* -8 for $ra and $fp, -32 for $a0-$a7, -8 for $v0 and 4-byte padding */
    emit_code (ast_func, "\tmove    $t9, %s\n", reg.name [reg1]);
    emit_code (ast_func, "\tjalr    $t9\n");
    emit_code (ast_func, "\tnop\n");
    emit_code (ast_func, "\tlw      $gp, %d($fp)\n", total_arg_size);

    reg2 = push_reg ();
    emit_code (ast_func, "\tmove    %s, $v0\n", reg.name [reg2]);
    emit_code (ast_func, "\tlw      $v0, %d($fp)\t# restore return value\n",
               frame_size - 48);
    /* -8 for $ra and $fp, -32 for $a0-$a7, -8 for $v0 and 4-byte padding */
}

/* stack layout:
  lower address
      <- 8 byte aligned
  □  argument 1-4 (also kept in $4-$7) <- fp, sp
  □  argument 5-n
      <- 8 byte aligned
  □  saved $gp
      <- 8 byte aligned
  □  local var 1
  □  local var 2-n
      <- 8 byte aligned
  □  tmp value ($v0)
      <- 8 byte aligned
  □  $s0
      ...
  □  $s7
  □  old $fp
  □  return address 
      <- 8 byte aligned
  □  <- old fp, sp */
static void
codegen_exp_id (struct AST *ast)
{
    int reg1;
    int offset;
    char *insn;
    struct Symbol *sym;

    sym = sym_lookup (ast->child [0]->u.id);
    assert (sym != NULL);

    switch (sym->type->kind) {
    case TYPE_KIND_PRIM:
        switch (sym->type->u.t_prim.ptype) {
        case PRIM_TYPE_CHAR: insn = "lb"; break;
        case PRIM_TYPE_INT:  insn = "lw"; break;
        case PRIM_TYPE_VOID: /* falling through */
        default:             assert (0); break;
        }
    case TYPE_KIND_POINTER:  insn = "lw"; break;
    case TYPE_KIND_FUNCTION: insn = "lw"; break;
    default:                 assert (0); break;
    }

    reg1 = push_reg ();
    switch (sym->name_space) {
    case NS_LOCAL:
        offset = total_arg_size + sym->offset + 8;
        /* 8 bytes for %gp + 4byte alignment */
	emit_code (ast, "\t%s      %s, %d($fp)\t# %s, %d\n",
                   insn, reg.name [reg1], offset, sym->name, sym->offset);
	break;
    case NS_ARG:
        offset = frame_size + sym->offset;
        /* $a0-$a3 are not used for simplicity.
           copying $a0-$a3 to stack is done in codegen_func */
	emit_code (ast, "\t%s      %s, %d($fp)\t# %s, %d\n",
                   insn, reg.name [reg1], offset, sym->name, sym->offset);
	break;
    case NS_GLOBAL:
	emit_code (ast, "\tlw      %s, %%got(%s)($gp)\n",
                   reg.name [reg1], sym->name);
	if (sym->type->kind != TYPE_KIND_FUNCTION) {
	    emit_code (ast, "\t%s      %s, 0(%s)\n",
                       insn, reg.name [reg1], reg.name [reg1]);
	}
	break;
    case NS_LABEL: /* falling through */
    default: assert (0); break;
    }
}

static void
codegen_exp (struct AST *ast)
{
    struct String *string;
    int reg1, reg2;
    char *insn;

    if (   !strcmp (ast->ast_type, "AST_expression_int")
        || !strcmp (ast->ast_type, "AST_expression_char")) {
	reg1 = push_reg ();
	emit_code (ast, "\tli      %s, %d\n", reg.name [reg1], ast->u.int_val);
    } else if (!strcmp (ast->ast_type, "AST_expression_string")) {
	string = string_lookup (ast->u.id);
        assert (string != NULL);
	reg1 = push_reg ();
	emit_code (ast, "\tlw      %s, %%got(%s%s)($gp) \t# \"%s\"\n",
                   reg.name [reg1], "$XCC.", string->label, string->data);
        emit_code (ast, "\taddiu   %s, %s, %%lo(%s%s)\n",
                   reg.name [reg1], reg.name [reg1], "$XCC.", string->label);
    } else if (!strcmp (ast->ast_type, "AST_expression_id")) {
	codegen_exp_id (ast);
    } else if (   !strcmp (ast->ast_type, "AST_expression_funcall1")
               || !strcmp (ast->ast_type, "AST_expression_funcall2")) {
	codegen_exp_funcall (ast);
    } else if (!strcmp (ast->ast_type, "AST_expression_assign")) {
	codegen_exp (ast->child [1]);
	codegen_lvalue (ast->child [0]);
	reg1 = pop_reg ();
	reg2 = top_reg ();

        switch (ast->child [0]->type->kind) {
        case TYPE_KIND_PRIM:
            switch (ast->child [0]->type->u.t_prim.ptype) {
            case PRIM_TYPE_CHAR: insn = "sb"; break;
            case PRIM_TYPE_INT:  insn = "sw"; break;
            case PRIM_TYPE_VOID: /* falling through */
            default:             assert (0); break;
            }
        case TYPE_KIND_POINTER:  insn = "sw"; break;
        case TYPE_KIND_FUNCTION: /* falling through */
        default:                 assert (0); break;
        }

	emit_code (ast, "\t%s      %s, 0(%s)\n",
                   insn, reg.name [reg2], reg.name [reg1]);
    } else if (!strcmp (ast->ast_type, "AST_expression_add")) {
	codegen_exp_add (ast);
    } else if (!strcmp (ast->ast_type, "AST_expression_sub")) {
	codegen_exp_sub (ast);
    } else if (   !strcmp (ast->ast_type, "AST_expression_lor")
               || !strcmp (ast->ast_type, "AST_expression_land")
               || !strcmp (ast->ast_type, "AST_expression_mul")
               || !strcmp (ast->ast_type, "AST_expression_div")) {
	codegen_exp_binop1 (ast);
    } else if (   !strcmp (ast->ast_type, "AST_expression_eq")
               || !strcmp (ast->ast_type, "AST_expression_less")) {
	codegen_exp_binop2 (ast);
    } else if (!strcmp (ast->ast_type, "AST_expression_unary")) {
	codegen_exp_unary (ast);
    } else if (!strcmp (ast->ast_type, "AST_expression_paren")) {
	codegen_exp (ast->child [0]);
    } else {
        assert (0);
    }
}

static char *
create_ctrl_label (void)
{
    static int num = 0;
    int size;
    char *label;
    size = snprintf (NULL, 0, "$XCC.CL%d", num);
    label = emalloc (size + 1);
    snprintf (label, size + 1, "$XCC.CL%d", num);
    num++;
    return label;
}

static void
codegen_stmt (struct AST *ast_stmt)
{
    int reg1;
    char *label1, *label2;
    struct Symbol *sym;

    if (!strcmp (ast_stmt->ast_type, "AST_statement_exp")) {
	if (!strcmp (ast_stmt->child [0]->ast_type, "AST_expression_opt_single")) {
	    codegen_exp (ast_stmt->child [0]->child [0]);
	    pop_reg ();
	} else if (!strcmp (ast_stmt->child [0]->ast_type, "AST_expression_opt_null")) {
            /* nothing to do */
        } else {
            assert (0);
        }
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_comp")) {
	codegen_block (ast_stmt->child [0]);
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_if")) {
	codegen_exp (ast_stmt->child [0]);
	reg1 = pop_reg ();
	label1 = create_ctrl_label ();
	emit_code (ast_stmt, "\tbeq     %s, $0, %s\n",
                   reg.name [reg1], label1);
	emit_code (ast_stmt, "\tnop\n");
	codegen_stmt (ast_stmt->child [1]);
	emit_code (ast_stmt, "%s:\n", label1);
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_ifelse")) {
	codegen_exp (ast_stmt->child [0]);
	reg1 = pop_reg ();
	label1 = create_ctrl_label ();
	label2 = create_ctrl_label ();
	emit_code (ast_stmt, "\tbeq     %s, $0, %s\n",
                   reg.name [reg1], label1);
	emit_code (ast_stmt, "\tnop\n");
	codegen_stmt (ast_stmt->child [1]);
	emit_code (ast_stmt, "\tj       %s\n", label2);
	emit_code (ast_stmt, "\tnop\n");
	emit_code (ast_stmt, "%s:\n", label1);
	codegen_stmt (ast_stmt->child [2]);
	emit_code (ast_stmt, "%s:\n", label2);
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_while")) {
	label1 = create_ctrl_label ();
	label2 = create_ctrl_label ();
	emit_code (ast_stmt, "%s:\n", label1);
	codegen_exp (ast_stmt->child [0]);
	reg1 = pop_reg ();
	emit_code (ast_stmt, "\tbeq     %s, $0, %s\n",
                   reg.name [reg1], label2);
	emit_code (ast_stmt, "\tnop\n");
	codegen_stmt (ast_stmt->child [1]);
	emit_code (ast_stmt, "\tj       %s\n", label1);
	emit_code (ast_stmt, "\tnop\n");
	emit_code (ast_stmt, "%s:\n", label2);
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_goto")) {
	label1 = ast_stmt->child [0]->u.id;
	sym = sym_lookup_label (label1);
	if (sym == NULL) {
	    fprintf (stderr, "No such label: %s\n", label1);
	    exit (1);
	}
	emit_code (ast_stmt, "\tj       $XCC.%s.%s:\n", func_name, label1);
	emit_code (ast_stmt, "\tnop\n");
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_label")) {
	emit_code(ast_stmt, "$XCC.%s.%s:\n",
                  func_name, ast_stmt->child [0]->u.id);
	codegen_stmt (ast_stmt->child [1]);
    } else if (!strcmp (ast_stmt->ast_type, "AST_statement_return")) {
	if (!strcmp (ast_stmt->child [0]->ast_type, "AST_expression_opt_single")) {
	    codegen_exp (ast_stmt->child [0]->child [0]);
	    reg1 = pop_reg ();
	    emit_code (ast_stmt, "\tmove    $v0, %s \t# return value\n",
                       reg.name [reg1]);
	}
	emit_code (ast_stmt, "\tj       $XCC.RE.%s\n", func_name);
	emit_code (ast_stmt, "\tnop\n");
    } else {
        assert (0);
    }
}

static void
codegen_begin_block (struct AST *ast)
{
    assert (!strcmp (ast->ast_type, "AST_compound_statement"));
    sym_table.local [++sym_table.local_index] = ast->u.local;
}

static void
codegen_end_block (void)
{
    sym_table.local_index--;
}

static void
codegen_block (struct AST *ast_block)
{
    struct AST *ast, *ast_stmt_list;
    assert (!strcmp (ast_block->ast_type, "AST_compound_statement"));
    codegen_begin_block (ast_block);
    
    ast_stmt_list = ast_block->child [1];
    ast = search_AST_bottom (ast_stmt_list, "AST_statement_list_single", NULL);
    while (1) {
        if (!strcmp (ast->ast_type, "AST_statement_list_single"))
            codegen_stmt (ast->child [0]); 
        else if (!strcmp (ast->ast_type, "AST_statement_list_pair"))
            codegen_stmt (ast->child [1]);
        else
            assert (0);
	if (ast == ast_stmt_list)
	    break;
	ast = ast->parent;
    } 
    codegen_end_block ();
}

static void
codegen_begin_function (struct AST *ast)
{
    assert(!strcmp (ast->ast_type, "AST_function_definition"));
    sym_table.local_index = -1;
    sym_table.global = ast->u.func.global;
    sym_table.arg    = ast->u.func.arg;
    sym_table.label  = ast->u.func.label;
    sym_table.string = ast->u.func.string;
}

static void
codegen_end_function (void)
{
    /* do nothing */
}

static void
codegen_func (struct AST *ast)
{
    int i;
    static int func_num = 0;
    struct String *string, *head;

    assert (!strcmp (ast->ast_type, "AST_function_definition"));

    codegen_begin_function (ast);

    /* string literals */
    head = sym_table.string;
    if (head != NULL)
	emit_code (ast, "\t.rdata\n");
    for (string = head; string != NULL; string = string->next) {
	emit_code (ast, "\t.align  2\n");
	emit_code (ast, "%s%s:\n", "$XCC.", string->label);
	emit_code (ast, "\t.ascii  \"%s\\000\"\n", string->data);
    }

    /* function prologue */
    func_name = ast->type->id;
    total_arg_size   = ast->u.func.total_arg_size;
    total_local_size = ast->u.func.total_local_size;
    frame_size = ROUNDUP_STACK (total_arg_size + total_local_size + 56);
    /* 16 bytes for saving $gp, $ra, $fp and 4-byte padding */
    /* 32 bytes for saving $s0-$s7 */
    /*  8 bytes for saving $v0 and 4-byte padding */

    emit_code (ast, "\t.text\n");
    emit_code (ast, "\t.align  2\n");
    emit_code (ast, "\t.globl  %s\n", func_name);
    emit_code (ast, "\t.ent    %s\n", func_name);
    emit_code (ast, "\t.type   %s, @function\n", func_name);

    emit_code (ast, "%s:\n", func_name);
    emit_code (ast, "\t.set    nomips16\n");
    emit_code (ast, "\t.frame  $fp,%d,$ra\n", frame_size);
    emit_code (ast, "\t.mask   0xc0ff0000,-4\n");
    emit_code (ast, "\t.fmask  0x00000000,0\n");
    emit_code (ast, "\t.set    noreorder\n");
    emit_code (ast, "\t.cpload $t9\n");
    emit_code (ast, "\t.set    nomacro\n");

    emit_code (ast, "\taddiu   $sp, $sp, %d\n", -frame_size);
    emit_code (ast, "\tsw      $ra, %d($sp)\n", frame_size - 4);
    emit_code (ast, "\tsw      $fp, %d($sp)\n", frame_size - 8);

    /* for simplicity, all callee-save registers are saved on stack */
    for (i = 0; i < 8; i++) {
        emit_code (ast, "\tsw      $s%d, %d($sp)\n",
                   i, frame_size - 8 - 32 + i*4);
        /* -8 for %ra, %fp, -32 for %s0-%s7 */
    }
    
    emit_code (ast, "\tmove    $fp, $sp\n");
    emit_code (ast, "\t.cprestore %d\n", total_arg_size);

    /* for simplicity, callee accesses arguments on stack, not on regs */
    for (i = 0; i < 4; i++) 
        emit_code (ast, "\tsw      $a%d, %d($fp)\n", i, frame_size + i*4);

    /* function body */
    codegen_block (ast->child [2]);

    /* function epilogue */
    emit_code (ast, "$XCC.RE.%s:\n", func_name);
    emit_code (ast, "\tmove    $sp, $fp\n");
    emit_code (ast, "\tlw      $ra, %d($sp)\n", frame_size - 4);
    emit_code (ast, "\tlw      $fp, %d($sp)\n", frame_size - 8);
    for (i = 0; i < 8; i++) {
        emit_code (ast, "\tlw      $s%d, %d($sp)\n",
                   i, frame_size - 8 - 32 + i*4);
        /* -8 for %ra, %fp, -32 for %s0-%s7 */
    }
    emit_code (ast, "\taddiu   $sp, $sp, %d\n", frame_size);
    emit_code (ast, "\tj       $ra\n");
    emit_code (ast, "\tnop\n");
    emit_code (ast, "\t.end    %s\n", func_name);
    func_num++;

    codegen_end_function ();
}

static void
codegen_dec (struct AST *ast)
{
    int size;
    char *id;
    assert (!strcmp (ast->ast_type, "AST_declaration"));

    size = ast->type->size;
    if (size <= 0)
	return;
    id   = ast->type->id;

    emit_code (ast, "\t.globl  %s\n", id);
    emit_code (ast, "\t.data\n");
    emit_code (ast, "\t.align  2\n");
    emit_code (ast, "\t.type   %s, @object\n", id);
    emit_code (ast, "\t.size   %s, %d\n", id, size);
    emit_code (ast, "%s:\n", id);
    emit_code (ast, "\t.skip   %d\n\n", size);
}
/* ---------------------------------------------------------------------- */
void
codegen (void)
{
    struct AST *ast, *ast_ext;
    ast = search_AST_bottom (ast_root, "AST_translation_unit_single", NULL);

    emit_code (ast_root, "\t.file   1 \"%s\"\n", filename);
    while (1) {
        if (!strcmp (ast->ast_type, "AST_translation_unit_single"))
            ast_ext = ast->child [0];
        else if (!strcmp (ast->ast_type, "AST_translation_unit_pair"))
            ast_ext = ast->child [1];
        else
            assert (0);

        if (!strcmp (ast_ext->ast_type, "AST_external_declaration_func"))
	    codegen_func (ast_ext->child [0]);
        else if (!strcmp (ast_ext->ast_type, "AST_external_declaration_dec"))
	    codegen_dec (ast_ext->child [0]);
        else 
            assert (0);

	if (ast == ast_root)
	    break;
	ast = ast->parent;
    }
    emit_code (ast_root, "\t.ident  \"XCC ver 0.1 ($Revision: 1.16 $)\"\n");
}
/* ---------------------------------------------------------------------- */

#if 0
レジスタ不足の際の対応．

#endif
